"""
osv to NVD CVE converter

This module fetches the vulnerability data from osv.dev and stores them in NVD CVE 1.1 json format.
"""

from zipfile import ZipFile

import httpx
import orjson
from semver import Version

from vdb.lib import CustomNamedTemporaryFile, config
from vdb.lib.nvd import NvdSource
from vdb.lib.utils import (
    compress_str,
    convert_score_severity,
    get_cvss3_from_vector,
    get_default_cve_data,
    parse_purl,
)


# Size of the stream to read and write to the file
DOWNLOAD_CHUNK_SIZE = 4096

vendor_overrides = {
    "apk": "alpine",
    "deb": "debian",
    "go": "golang",
    "crates.io": "cargo",
    "swifturl": "swift",
    "github actions": "github",
}


def functions_to_modules(affected_functions):
    affected_modules = set()
    for af in affected_functions:
        if "::" in af:
            tmp_parts = af.split("::")
            affected_modules.add("::".join(tmp_parts[:-1]))
        elif "/" in af:
            tmp_parts = af.split("/")
            affected_modules.add("/".join(tmp_parts[:-1]))
    return affected_modules


class OSVSource(NvdSource):
    """OSV CVE source"""

    def download_all(self):
        """Download all cve data"""
        # For performance do not retain the whole data in-memory
        # See: https://github.com/AppThreat/vulnerability-db/issues/27
        data_list = []
        for _, url in config.osv_url_dict.items():
            data = self.fetch(url)
            if data:
                self.store(data)
        return data_list

    def download_recent(self):
        pass

    def fetch(self, url):
        ret_data = []
        client = httpx.Client(http2=True, follow_redirects=True, timeout=180)
        with CustomNamedTemporaryFile() as tf:
            try:
                with client.stream("GET", url) as r:
                    for chunk in r.iter_bytes(chunk_size=DOWNLOAD_CHUNK_SIZE):
                        tf.write(chunk)
                    tf.flush()
            except Exception:
                return ret_data
            with ZipFile(tf.name, "r") as zipfp:
                for zf in zipfp.namelist():
                    if zf.endswith(".json"):
                        with zipfp.open(zf) as jsonf:
                            cve_data = jsonf.read()
                            try:
                                json_data = orjson.loads(cve_data)
                                ret_data += self.convert(json_data)
                            except Exception:
                                pass
        return ret_data

    def convert(self, cve_data):
        if cve_data.get("withdrawn"):
            return []
        return self.to_vuln(cve_data)

    @staticmethod
    def to_vuln(cve_data):
        ret_data = []
        if not cve_data.get("affected"):
            return ret_data
        cve_id = cve_data.get("id")
        cwe_id = ""
        edition = "*"
        cve_references = cve_data.get("references", [])
        aliases = cve_data.get("aliases", [])
        aliases_block = ""
        if aliases and len(aliases) > 1:
            aliases_block = f"""
## Related CVE(s)
{", ".join(aliases)}
"""
        description = f"""# {cve_data.get("summary", "Summary")}
{cve_data.get("details", "")}
{aliases_block}
"""
        if "** DISPUTED **" in description or "** REJECT **" in description:
            return ret_data
        references = []
        # Change the key from type to name in references
        for aref in cve_references:
            references.append({"name": aref.get("type", "url"), "url": aref.get("url")})
        references = orjson.dumps(references, option=orjson.OPT_NAIVE_UTC)
        if isinstance(references, bytes):
            references = references.decode("utf-8", "ignore")
        # Quality of PYSEC data is quite low missing both severity and score
        # Where a PYSEC feed also reference a github id, let's ignore it since G comes before P
        #   so it would have gotten processed
        # Fixes #99
        if cve_id.startswith("PYSEC"):
            for i in aliases:
                if i.startswith("GHSA"):
                    return ret_data
        # Try to locate the CVE id from the aliases section
        if not cve_id.startswith("CVE") and not cve_id.startswith("RUSTSEC"):
            for i in aliases:
                if i.startswith("CVE"):
                    cve_id = i
                    break
        assigner = "google"
        vector_string = ""
        if cve_id.startswith("GHSA"):
            assigner = "github_m"
        elif cve_id.startswith("CVE"):
            assigner = "mitre"
        elif cve_id.startswith("NPM"):
            assigner = "github_m"
        # For malwares, default to critical
        severity = "CRITICAL" if cve_id.startswith("MAL") else "LOW"
        if cve_data.get("severity"):
            severity_list = cve_data.get("severity")
            for sv in severity_list:
                if sv["type"] == "CVSS_V3":
                    vector_string = sv["score"]
        # Issue 58
        cve_database_specific = cve_data.get("database_specific")
        cve_ecosystem_specific = cve_data.get("ecosystem_specific")
        # Support for collecting affected functions
        affected_functions = set()
        affected_modules = set()
        if cve_database_specific:
            if cve_database_specific.get("severity"):
                severity = cve_database_specific.get("severity")
            if cve_database_specific.get("cwe_ids"):
                cwes = cve_database_specific.get("cwe_ids")
                if isinstance(cwes, list):
                    cwe_id = ",".join(cwes)
        if cve_ecosystem_specific and cve_ecosystem_specific.get("severity"):
            severity = cve_ecosystem_specific.get("severity")
        for pkg_data in cve_data.get("affected"):
            if pkg_data.get("ecosystem_specific"):
                ecosystem_specific = pkg_data.get("ecosystem_specific")
                if ecosystem_specific.get("severity"):
                    severity = ecosystem_specific.get("severity")
                if ecosystem_specific.get("affected_functions"):
                    affected_functions = affected_functions.union(ecosystem_specific.get("affected_functions"))
                if ecosystem_specific.get("affects", {}).get("functions"):
                    affected_functions = affected_functions.union(ecosystem_specific.get("affects").get("functions"))
                for aimp in ecosystem_specific.get("imports", []):
                    if aimp.get("path"):
                        affected_modules.add(aimp.get("path"))
                    if aimp.get("symbols"):
                        affected_functions = affected_functions.union(aimp.get("symbols"))
            if pkg_data.get("database_specific"):
                database_specific = pkg_data.get("database_specific")
                if database_specific.get("cwes"):
                    cwes = database_specific.get("cwes")
                    if isinstance(cwes, list):
                        cwe_id = cwes[0].get("cweId")
                if database_specific.get("cvss"):
                    cvss = database_specific.get("cvss")
                    if isinstance(cvss, dict):
                        if cvss.get("severity"):
                            severity = cvss.get("severity", "").upper()
                        if not vector_string and cvss.get("vectorString"):
                            vector_string = cvss.get("vectorString")
                        if cvss.get("score"):
                            score = cvss.get("score")
                            severity = convert_score_severity(score)
            user_interaction = "REQUIRED"
            if vector_string:
                cvss3_obj = get_cvss3_from_vector(vector_string)
                score = cvss3_obj.get("baseScore")
                severity = cvss3_obj.get("baseSeverity")
                exploitability_score = cvss3_obj.get("temporalScore")
                attack_complexity = cvss3_obj.get("attackComplexity")
                user_interaction = cvss3_obj.get("userInteraction")
            else:
                (
                    score,
                    severity,
                    dvector_string,
                    attack_complexity,
                ) = get_default_cve_data(severity)
                # Override the score for malware
                if cve_id.startswith("MAL"):
                    score = 10.0
                    user_interaction = "NONE"
                # Set the default vector string only if unavailable
                if not vector_string and dvector_string:
                    vector_string = dvector_string
                exploitability_score = score
            ranges = pkg_data.get("ranges", [])
            versions_list = pkg_data.get("versions", [])
            vendor_ecosystem = pkg_data.get("package", {}).get("ecosystem", "").lower()
            vendor = vendor_ecosystem
            pkg_name = pkg_data.get("package", {}).get("name", "")
            pkg_name_list = []
            purl = parse_purl(pkg_data.get("package", {}).get("purl", ""))
            if purl:
                if purl.get("type"):
                    vendor = purl["type"]
                if purl.get("namespace") and purl.get("name"):
                    pkg_name = f'{purl["namespace"]}/{purl["name"]}'
                elif purl.get("name"):
                    pkg_name = purl["name"]
            if ":" in pkg_name and vendor.lower() not in (
                "swift",
                "swifturl",
                "github",
                "github actions",
            ):
                # Example: commons-fileupload:commons-fileupload
                # org.apache.tomcat:tomcat
                tmp_a = pkg_name.split(":")
                if len(tmp_a) == 2:
                    pkg_name = f"{tmp_a[0]}/{tmp_a[-1]}"
            # In case of swift, we need to remove any https protocol from the name
            if pkg_name.startswith("https://"):
                pkg_name = pkg_name.removeprefix("https://")
            if pkg_name:
                pkg_name_list.append(pkg_name)
            # Substitute alpine for apk and debian for deb
            if vendor_overrides.get(vendor):
                vendor = vendor_overrides.get(vendor)
            # Since swift allows both url and local based lookups, we store both the variations
            if vendor in (
                "swift",
                "swifturl",
                "github",
                "github actions",
            ) and pkg_name.startswith("github.com"):
                pkg_name_list.append(pkg_name.removeprefix("github.com/"))
            # For OS packages, such as alpine OSV appends the os version to the vendor
            # Let's remove it and add it to package name
            if ":" in vendor_ecosystem and (
                "alpine" in vendor
                or "apk" in vendor
                or "deb" in vendor
                or "debian" in vendor
                or "almalinux" in vendor
                or "rocky" in vendor
            ):
                tmp_v = vendor_ecosystem.split(":")
                vendor = tmp_v[0].lower().replace(" ", "").replace("-", "")
                vdistro = tmp_v[1]
                if vendor in ("alpine", "apk"):
                    vdistro = vdistro.replace("v", "")
                # In os-release, ID for rockylinux is simply rocky
                if "rocky" in vendor:
                    vendor = vendor.replace("linux", "")
                edition = f"{vendor}-{vdistro}"
                # Only use the precise version for os packages
                if (
                    "debian" in vendor
                    or "deb" in vendor
                    or "alpine" in vendor
                    or "apk" in vendor
                    or "almalinux" in vendor
                    or "rocky" in vendor
                ):
                    pkg_name_list = [
                        f"{vendor}/{edition}/{pkg_name.removeprefix(vendor + '/')}"
                    ]
                else:
                    pkg_name_list.append(f"{edition}/{pkg_name}")
            # For some ecosystem, osv provides a full list of versions with partial events. See osv-pypi2.json for an example
            # Problem 1: Storing each version as-is is slowing down vdb
            # Problem 2: The versions_list may be unsorted and trying to sort based on semantic versions can fail
            # Solution: We do our best to sort the versions_list. If it fails, we assume the input is sorted and store the first and last entry as min and max.
            # The assumption seems to be working as of today, but could result in false positives or false negatives in the future.
            needs_version_backup = len(ranges) == 0
            possible_fix_version = ""
            for r in ranges:
                events = r.get("events")
                for ev in events:
                    if ev.get("introduced", "") in (0, "0", "0.0.0"):
                        needs_version_backup = True
                    if ev.get("fixed"):
                        possible_fix_version = ev.get("fixed")
            if needs_version_backup and len(versions_list):
                try:
                    min_ver = min(
                        versions_list,
                        key=lambda x: Version.parse(x, optional_minor_and_patch=True),
                    )
                    max_ver = max(
                        versions_list,
                        key=lambda x: Version.parse(x, optional_minor_and_patch=True),
                    )
                except Exception:
                    min_ver = versions_list[0]
                    max_ver = versions_list[-1]
                for full_pkg in pkg_name_list:
                    tdata = config.CVE_TPL % dict(
                        cve_id=cve_id,
                        cwe_id=cwe_id,
                        assigner=assigner,
                        references=references,
                        description="",
                        vectorString=vector_string,
                        vendor=vendor,
                        product=full_pkg,
                        version="*",
                        edition=edition,
                        version_start_including=min_ver,
                        version_end_including=max_ver,
                        version_start_excluding="",
                        version_end_excluding="",
                        fix_version_start_including=possible_fix_version,
                        fix_version_end_including="",
                        fix_version_start_excluding=max_ver,
                        fix_version_end_excluding="",
                        severity=severity,
                        attackComplexity=attack_complexity,
                        score=score,
                        userInteraction=user_interaction,
                        exploitabilityScore=exploitability_score,
                        publishedDate=cve_data.get("published"),
                        lastModifiedDate=cve_data.get("modified"),
                    )
                    try:
                        vuln = NvdSource.convert_vuln(orjson.loads(tdata))
                        if vuln is None:
                            continue
                        vuln.description = compress_str(description)
                        if affected_functions:
                            vuln.affects = {
                                "affected_functions": sorted(list(affected_functions)),
                                "affected_modules": sorted(
                                    list(
                                        affected_modules
                                        | functions_to_modules(affected_functions)
                                    )
                                ),
                            }
                        ret_data.append(vuln)
                    except Exception:
                        pass
            for r in ranges:
                if r.get("type") == "GIT" and r.get("repo"):
                    vendor = "generic"
                    repo_name = (
                        r.get("repo")
                        .removeprefix("http://")
                        .removeprefix("https://")
                        .removeprefix("git://")
                        .removesuffix("/")
                        .removesuffix(".git")
                        .lower()
                    )
                    # See #112
                    for special_type in ("github.com", "gitlab.com"):
                        if repo_name.startswith(special_type):
                            vendor = special_type.removesuffix(".com")
                            repo_name = repo_name.replace(f"{special_type}/", "")
                    pkg_name_list.append(repo_name)
                events = r.get("events")
                rversions_list = r.get("versions", [])
                version_end_including = ""
                version_start_excluding = ""
                version_end_excluding = ""
                fix_version_start_including = ""
                fix_version_end_including = ""
                fix_version_start_excluding = ""
                fix_version_end_excluding = ""
                version_start_including = ""
                if rversions_list:
                    version_start_including = rversions_list[0]
                    if (
                        len(rversions_list) > 1
                        and version_start_including != rversions_list[-1]
                    ):
                        version_end_including = rversions_list[-1]
                for ev in events:
                    # Reset all versions for introduced event
                    if ev.get("introduced") is not None:
                        version_end_including = ""
                        version_start_excluding = ""
                        version_end_excluding = ""
                        fix_version_start_including = ""
                        fix_version_end_including = ""
                        fix_version_start_excluding = ""
                        fix_version_end_excluding = ""
                        version_start_including = ev.get("introduced").split(":")[-1]
                        # Sometimes, the quality of data will be so low we need
                        # workarounds like this to make the result searchable
                        if version_start_including in (0, "0", "0.0.0"):
                            version_start_including = "0.0.0"
                    if ev.get("fixed"):
                        fix_version_start_including = ev.get("fixed").split(":")[-1]
                        fix_version_end_including = ev.get("fixed").split(":")[-1]
                        version_end_excluding = ev.get("fixed").split(":")[-1]
                    if ev.get("last_affected"):
                        version_end_including = ev.get("last_affected").split(":")[-1]
                        fix_version_start_excluding = version_end_including
                    if ev.get("limit"):
                        version_end_excluding = ev.get("limit").split(":")[-1]
                    # Create an entry for each introduced + fixed/limit event
                    if version_start_including and (
                        fix_version_start_including
                        or version_end_including
                        or version_end_excluding
                        or (len(events) == 1 and not versions_list)
                    ):
                        for full_pkg in pkg_name_list:
                            tdata = config.CVE_TPL % dict(
                                cve_id=cve_id,
                                cwe_id=cwe_id,
                                assigner=assigner,
                                references=references,
                                description="",
                                vectorString=vector_string,
                                vendor=vendor,
                                product=full_pkg,
                                version="*",
                                edition=edition,
                                version_start_including=version_start_including,
                                version_end_including=version_end_including,
                                version_start_excluding=version_start_excluding,
                                version_end_excluding=version_end_excluding,
                                fix_version_start_including=fix_version_start_including,
                                fix_version_end_including=fix_version_end_including,
                                fix_version_start_excluding=fix_version_start_excluding,
                                fix_version_end_excluding=fix_version_end_excluding,
                                severity=severity,
                                attackComplexity=attack_complexity,
                                score=score,
                                userInteraction=user_interaction,
                                exploitabilityScore=exploitability_score,
                                publishedDate=cve_data.get("published"),
                                lastModifiedDate=cve_data.get("modified"),
                            )
                            try:
                                vuln = NvdSource.convert_vuln(orjson.loads(tdata))
                                if vuln is None:
                                    continue
                                vuln.description = compress_str(description)
                                if affected_functions:
                                    vuln.affects = {
                                        "affected_functions": sorted(
                                            list(affected_functions)
                                        ),
                                        "affected_modules": sorted(
                                            list(
                                                affected_modules
                                                | functions_to_modules(
                                                    affected_functions
                                                )
                                            )
                                        ),
                                    }
                                ret_data.append(vuln)
                            except Exception:
                                pass
        return ret_data



import os
import re
import tempfile
from abc import ABCMeta, abstractmethod
from datetime import datetime
from enum import Enum

import orjson

# Known application package types
# See https://github.com/package-url/purl-spec/blob/master/PURL-TYPES.rst
# chainguard and wolfi has been added for suppression purposes since the data quality is poor
KNOWN_PKG_TYPES = [
    "alpm",
    "bitbucket",
    "bitnami",
    "cargo",
    "composer",
    "cocoapods",
    "conda",
    "cpan",
    "cran",
    "docker",
    "generic",
    "github",
    "gitlab",
    "huggingface",
    "mlflow",
    "qpkg",
    "oci",
    "maven",
    "npm",
    "nuget",
    "pypi",
    "gem",
    "rubygems",
    "golang",
    "clojars",
    "conan",
    "pub",
    "hackage",
    "android",
    "dwf",
    "gsd",
    "hex",
    "packagist",
    "uvi",
    "apk",
    "deb",
    "rpm",
    "linux",
    "swid",
    "oss-fuzz",
    "ebuild",
    "swift",
    "apache",
    "android",
    "atom",
    "bower",
    "brew",
    "buildroot",
    "carthage",
    "chef",
    "chocolatey",
    "coreos",
    "ctan",
    "crystal",
    "drupal",
    "dtype",
    "dub",
    "elm",
    "eclipse",
    "gitea",
    "gradle",
    "guix",
    "haxe",
    "helm",
    "julia",
    "lua",
    "melpa",
    "meteor",
    "nim",
    "nix",
    "opam",
    "openwrt",
    "osgi",
    "p2",
    "pear",
    "pecl",
    "perl6",
    "platformio",
    "puppet",
    "sourceforge",
    "sublime",
    "terraform",
    "vagrant",
    "vim",
    "wordpress",
    "yocto"
]

# Maps variations of string to package types
PKG_TYPES_MAP = {
    "composer": ["php", "laravel", "wordpress", "joomla"],
    "maven": ["jenkins", "java", "kotlin", "groovy", "clojars", "hackage"],
    "npm": ["javascript", "node.js", "nodejs", "npmjs"],
    "nuget": [".net_framework", "csharp", ".net_core", "asp.net", "c#"],
    "pypi": ["python"],
    "gem": ["ruby"],
    "rubygems": ["ruby", "gem"],
    "golang": ["go"],
    "cargo": ["rust", "crates.io", "crates"],
    "pub": ["dart"],
    "hex": ["elixir"],
    "github": ["actions"],
    "apk": ["alpine"],
    "deb": ["ubuntu", "debian", "mint", "popos"],
    "rpm": [
        "redhat",
        "centos",
        "alma",
        "amazon",
        "rocky",
        "suse",
        "opensuse",
        "fedora",
        "fedoraproject"
    ],
    "alpm": ["arch", "archlinux"],
    "ebuild": ["gentoo", "portage"],
    "ios": ["iphoneos", "iphone_os"],
    "vscode": ["vs_code", "visual_studio_code"],
    "macos": ["mac_os_x", "osx"]
}

# CPE Regex
CPE_REGEX = re.compile(
    "cpe:?:[^:]+:[^:]+:(?P<vendor>[^:]+):(?P<package>[^:]+):(?P<version>[^:]+)?"
)

# CPE Full Regex including unused parameters
CPE_FULL_REGEX = re.compile(
    "cpe:?:[^:]+:(?P<cve_type>[^:]+):(?P<vendor>[^:]+):(?P<package>[^:]+):(?P<version>[^:]+):(?P<update>[^:]+):(?P<edition>[^:]+):(?P<lang>[^:]+):(?P<sw_edition>[^:]+):(?P<target_sw>[^:]+):(?P<target_hw>[^:]+):(?P<other>[^:]+)"
)


class VulnerabilitySource(metaclass=ABCMeta):
    @classmethod
    @abstractmethod
    def download_all(cls):
        pass

    @classmethod
    @abstractmethod
    def download_recent(cls):
        pass

    @classmethod
    @abstractmethod
    def bulk_search(cls, app_info, pkg_list):
        pass

    @classmethod
    @abstractmethod
    def convert(cls, data):
        pass

    @classmethod
    @abstractmethod
    def convert5(cls, data):
        pass

    @classmethod
    @abstractmethod
    def store(cls, data):
        pass

    @classmethod
    @abstractmethod
    def refresh(cls):
        pass


class Severity(str, Enum):
    UNSPECIFIED: str = "UNSPECIFIED"
    LOW: str = "LOW"
    MEDIUM: str = "MEDIUM"
    HIGH: str = "HIGH"
    CRITICAL: str = "CRITICAL"

    @staticmethod
    def from_str(sevstr):
        if isinstance(sevstr, dict):
            sevstr = sevstr["value"]
        if not sevstr:
            return Severity.UNSPECIFIED
        for k, v in Severity.__members__.items():
            if k == sevstr.upper():
                return v
        return Severity.UNSPECIFIED

    def __str__(self):
        return self.value


def convert_time(time_str: str) -> datetime | None:
    """Converts a variety of datetime formats to datetime object

    :param time_str: String time to convert
    :return: datetime object
    """
    if time_str is None or time_str == "":
        return None
    dt = None
    try:
        dt = datetime.fromisoformat(time_str)
    except ValueError:
        date_formats = [
            "%Y-%m-%dT%H:%M:%S",
            "%Y-%m-%dT%H:%M:%SZ",
            "%Y-%m-%dT%H:%Mz",
            "%Y-%m-%dT%H:%M:%S.%fZ",
        ]
        for fmt in date_formats:
            try:
                dt = datetime.strptime(time_str, fmt)
                break
            except ValueError:
                continue
    except TypeError:
        if isinstance(time_str, dict):
            time_str = time_str["value"]
            return convert_time(time_str)
    return dt or time_str


class CvssV3:
    """CVSS v3 representation"""

    base_score: float
    exploitability_score: float
    impact_score: float
    attack_vector: str
    attack_complexity: str
    privileges_required: str
    user_interaction: str
    scope: str
    confidentiality_impact: str
    integrity_impact: str
    availability_impact: str
    vector_string: str

    def __init__(
            self,
            base_score,
            exploitability_score,
            impact_score,
            attack_vector,
            attack_complexity,
            privileges_required,
            user_interaction,
            scope,
            confidentiality_impact,
            integrity_impact,
            availability_impact,
            vector_string=None,
    ):
        self.base_score = base_score
        self.exploitability_score = exploitability_score
        self.impact_score = impact_score
        self.attack_vector = attack_vector
        self.attack_complexity = attack_complexity
        self.privileges_required = privileges_required
        self.user_interaction = user_interaction
        self.scope = scope
        self.confidentiality_impact = confidentiality_impact
        self.integrity_impact = integrity_impact
        self.availability_impact = availability_impact
        self.vector_string = vector_string

    def to_dict(self):
        return {
            "base_score": self.base_score,
            "exploitability_score": self.exploitability_score,
            "impact_score": self.impact_score,
            "attack_vector": self.attack_vector,
            "attack_complexity": self.attack_complexity,
            "privileges_required": self.privileges_required,
            "user_interaction": self.user_interaction,
            "scope": self.scope,
            "confidentiality_impact": self.confidentiality_impact,
            "integrity_impact": self.integrity_impact,
            "availability_impact": self.availability_impact,
            "vector_string": self.vector_string,
        }

    def __repr__(self):
        return orjson.dumps(self.to_dict()).decode("utf-8", "ignore")


class PackageIssue:
    """Package issue class"""

    def __init__(
            self,
            affected_location,
            fixed_location,
            mii=None,
            mai=None,
            mie=None,
            mae=None,
    ):
        self.affected_location = VulnerabilityLocation.from_values(
            affected_location,
            mii,
            mai,
            mie,
            mae,
        )
        # If there is no fixed_location but there is max excluded version then consider that as the fix
        if not fixed_location and mae:
            self.fixed_location = mae
        else:
            if fixed_location and fixed_location.startswith("cpe"):
                # Extract the fixed version from fixed_location cpe uri
                fixed_parts = (
                    CPE_REGEX.match(fixed_location) if fixed_location else None
                )
                if fixed_parts:
                    self.fixed_location = fixed_parts.group("version")
                else:
                    self.fixed_location = ""
            else:
                self.fixed_location = fixed_location

    @staticmethod
    def from_dict(package_issue):
        return PackageIssue(
            package_issue.get("affected_location"), package_issue.get("fixed_location")
        )

    def to_dict(self) -> dict:
        return {
            "affected_location": self.affected_location.to_dict(),
            "fixed_location": self.fixed_location,
        }

    def __str__(self):
        return orjson.dumps(
            {
                "affected_location": str(self.affected_location),
                "fixed_location": self.fixed_location,
            },
            option=orjson.OPT_NAIVE_UTC,
        ).decode("utf-8", "ignore")


class VulnerabilityDetail:
    """Vulnerability detail class"""

    def __init__(
            self,
            cpe_uri: str,
            package: str,
            min_affected_version_including: str | None,
            max_affected_version_including: str | None,
            min_affected_version_excluding: str | None,
            max_affected_version_excluding: str | None,
            severity: str,
            description: str | None,
            fixed_location: str,
            package_type: str,
            is_obsolete: str,
            source_update_time: str,
    ):
        parts = CPE_REGEX.match(cpe_uri)
        self.cpe_uri = cpe_uri
        # Occasionally, NVD CPE value could be invalid. We need to guard against this
        if parts:
            self.package = package if package else parts.group("package")
            self.mii = (
                min_affected_version_including
                if min_affected_version_including
                else parts.group("version")
            )
            self.mai = (
                max_affected_version_including
                if max_affected_version_including
                else parts.group("version")
            )
        else:
            # Use split to extract the package name in case of bad CPE value
            package_workaround = ""
            if cpe_uri:
                cpe_parts = cpe_uri.split(":")
                if len(cpe_parts) > 4:
                    package_workaround = cpe_parts[4]
            self.package = package if package else package_workaround
            self.mii = (
                min_affected_version_including
                if min_affected_version_including
                else "*"
            )
            self.mai = (
                max_affected_version_including
                if max_affected_version_including
                else "*"
            )
        self.mie = (
            min_affected_version_excluding if min_affected_version_excluding else None
        )
        self.mae = (
            max_affected_version_excluding if max_affected_version_excluding else None
        )
        self.severity = Severity.from_str(severity)
        self.description = description
        self.fixed_location = fixed_location
        self.package_type = VulnerabilityDetail.get_type(cpe_uri, package_type)
        self.is_obsolete = is_obsolete
        self.source_update_time: datetime = convert_time(source_update_time)

    @staticmethod
    def get_type(cpe_uri, package_type):
        if package_type and package_type in KNOWN_PKG_TYPES:
            return package_type
        parts = CPE_REGEX.match(cpe_uri)
        # cpe:2.3:a:netaddr_project:netaddr:*:*:*:*:*:ruby:*:*
        all_parts = CPE_FULL_REGEX.match(cpe_uri)
        if all_parts:
            cve_type = all_parts.group("cve_type")
            if cve_type != "a":
                return cve_type
        if parts:
            ptype = parts.group("vendor")
            if ptype in KNOWN_PKG_TYPES:
                return ptype
            if all_parts and (
                    all_parts.group("target_sw") and all_parts.group("target_sw") != "*"
                    or (
                        all_parts.group("sw_edition")
                        and all_parts.group("sw_edition") != "*"
                    )
            ):
                for vk, vv in PKG_TYPES_MAP.items():
                    target_sw = all_parts.group("target_sw")
                    sw_edition = all_parts.group("sw_edition")
                    if vk in (target_sw, sw_edition):
                        return vk
                    if target_sw in vv or sw_edition in vv:
                        return vk
                return ptype
            # Unknown type. Just pass-through for now
            return ptype
        return None

    @staticmethod
    def from_dict(detail):
        return VulnerabilityDetail(
            detail.get("cpe_uri"),
            detail.get("package"),
            detail.get("mii"),
            detail.get("mai"),
            detail.get("mie"),
            detail.get("mae"),
            detail.get("severity"),
            detail.get("description"),
            detail.get("fixed_location"),
            detail.get("package_type"),
            detail.get("is_obsolete"),
            detail.get("source_update_time"),
        )

    def to_dict(self):
        return {
            "cpe_uri": self.cpe_uri,
            "package": self.package,
            "mii": self.mii,
            "mai": self.mai,
            "mie": self.mie,
            "mae": self.mae,
            "severity": self.severity.value,
            "description": self.description,
            "fixed_location": self.fixed_location,
            "package_type": self.package_type,
            "is_obsolete": self.is_obsolete,
            "source_update_time": self.source_update_time.strftime("%Y-%m-%dT%H:%M:%S")
            if isinstance(self.source_update_time, datetime)
            else self.source_update_time,
        }

    def __repr__(self):
        return orjson.dumps(self.to_dict()).decode("utf-8", "ignore")


class Vulnerability:
    """Vulnerability"""

    cvss_v3: CvssV3

    def __init__(
            self,
            vid: str,
            assigner: str,
            problem_type: str,
            score: float,
            severity: str,
            description: str,
            related_urls: list[str],
            details: list[VulnerabilityDetail],
            cvss_v3: CvssV3,
            source_update_time: str,
            source_orig_time: str,
            affects: dict = None,
    ):
        self.id = vid
        self.assigner = assigner
        self.problem_type = problem_type
        self.score = score
        self.severity = Severity.from_str(severity)
        self.description = description
        self.related_urls = related_urls
        self.details = details
        self.cvss_v3 = cvss_v3
        self.source_update_time: datetime = convert_time(source_update_time)
        self.source_orig_time: datetime = convert_time(source_orig_time)
        self.affects = affects

    def __repr__(self):
        return orjson.dumps(
            {
                "id": self.id,
                "assigner": self.assigner,
                "problem_type": self.problem_type,
                "score": self.score,
                "severity": self.severity.value,
                "description": self.description,
                "related_urls": self.related_urls,
                "details": str(self.details),
                "cvss_v3": str(self.cvss_v3),
                "source_update_time": self.source_update_time.strftime(
                    "%Y-%m-%dT%H:%M:%S"
                ),
                "source_orig_time": self.source_orig_time.strftime("%Y-%m-%dT%H:%M:%S"),
                "affects": self.affects
            },
            option=orjson.OPT_NAIVE_UTC,
        ).decode("utf-8", "ignore")


class VulnerabilityLocation:
    cpe_uri: str
    vendor: str
    package: str
    version: str

    def __init__(self, cpe_uri, vendor, package, version):
        self.cpe_uri = cpe_uri
        self.vendor = vendor
        self.package = package
        self.version = version

    @staticmethod
    def from_values(
            cpe_uri,
            mii=None,
            mai=None,
            mie=None,
            mae=None,
    ):
        parts = None
        version = "*"
        if not cpe_uri and not mii and not mai and not mie and not mae:
            return None
        if cpe_uri:
            parts = CPE_REGEX.match(cpe_uri)
            version = mai if mai else parts.group(3)
        version_left = ""
        version_right = ""
        if mie:
            version_left = ">" + mie
        if mii and mii != "*":
            version_left = ">=" + mii
        if mae:
            version_right = "<" + mae
        if mai and mai != "*":
            version_right = "<=" + mai
        if version_left and not version_right:
            version = version_left
            # Convert >0.0.0 to *
            if version == ">0.0.0":
                version = "*"
        elif not version_left and version_right:
            version = version_right
        elif version_left and version_right:
            if mii == mai:
                version = mai
            else:
                version = f"{version_left}-{version_right}"
        if parts:
            return VulnerabilityLocation(
                cpe_uri, parts.group("vendor"), parts.group("package"), version
            )
        return None

    def __str__(self):
        return orjson.dumps(
            {
                "cpe_uri": str(self.cpe_uri),
                "package": str(self.package),
                "version": str(self.version),
                "vendor": str(self.vendor),
            },
            option=orjson.OPT_NAIVE_UTC,
        ).decode("utf-8", "ignore")

    def to_dict(self):
        return {
            "cpe_uri": str(self.cpe_uri),
            "package": str(self.package),
            "version": str(self.version),
            "vendor": str(self.vendor),
        }


class VulnerabilityOccurrence:
    """Class to represent an occurrence of a vulnerability"""

    id: str
    problem_type: str
    type: str
    severity: Severity
    cvss_score: str
    cvss_v3: CvssV3
    package_issue: PackageIssue
    short_description: str
    long_description: str
    related_urls: list
    effective_severity: Severity
    source_update_time: datetime
    source_orig_time: datetime
    matched_by: str

    def __init__(
            self,
            oid,
            problem_type,
            otype,
            severity,
            cvss_score,
            cvss_v3,
            package_issue,
            short_description,
            long_description,
            related_urls,
            effective_severity,
            source_update_time,
            source_orig_time,
            matched_by,
    ):
        self.id = oid
        self.problem_type = problem_type
        self.type = otype
        self.severity = severity
        self.cvss_score = cvss_score
        self.cvss_v3 = cvss_v3 if cvss_v3 else None
        self.package_issue = package_issue
        self.short_description = short_description
        self.long_description = long_description
        self.related_urls = related_urls
        self.effective_severity = effective_severity
        self.source_update_time = source_update_time
        self.source_orig_time = source_orig_time
        self.matched_by = matched_by

    def to_dict(self):
        """Convert the object to dict"""
        if isinstance(self.source_update_time, datetime):
            source_update_time = self.source_update_time.strftime("%Y-%m-%dT%H:%M:%S")
        else:
            new_time = self.source_update_time
            source_update_time = (
                new_time.strftime("%Y-%m-%dT%H:%M:%S")
                if (isinstance(new_time, datetime))
                else None
            )
        if isinstance(self.source_orig_time, datetime):
            source_orig_time = self.source_orig_time.strftime("%Y-%m-%dT%H:%M:%S")
        else:
            new_time = self.source_orig_time
            source_orig_time = (
                new_time.strftime("%Y-%m-%dT%H:%M:%S")
                if (isinstance(new_time, datetime))
                else None
            )
        return {
            "id": self.id,
            "problem_type": self.problem_type
            if isinstance(self.problem_type, str)
            else ",".join(self.problem_type),
            "type": self.type,
            "severity": str(self.severity),
            "cvss_score": str(self.cvss_score),
            "cvss_v3": self.cvss_v3.to_dict() if self.cvss_v3 else {},
            "package_issue": self.package_issue.to_dict() if self.package_issue else {},
            "short_description": self.short_description,
            "long_description": self.long_description,
            "related_urls": self.related_urls,
            "effective_severity": str(self.effective_severity),
            "source_update_time": source_update_time,
            "source_orig_time": source_orig_time,
            "matched_by": self.matched_by,
        }


class CustomNamedTemporaryFile:
    def __init__(self, mode="wb", delete=True):
        self._mode = mode
        self._delete = delete
        self._temp_file = None

    def __enter__(self):
        # Generate a random temporary file name
        file_name = os.path.join(tempfile.gettempdir(), os.urandom(24).hex())
        # Ensure the file is created
        open(file_name, "x").close()
        # Open the file in the given mode
        self._temp_file = open(file_name, self._mode)
        return self._temp_file

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._temp_file:
            self._temp_file.close()
            if self._delete:
                os.remove(self._temp_file.name)



# pylint: disable=C0115, C0103, C0301

from enum import Enum
from typing import Annotated

from pydantic import Field, RootModel


class AttackVectorType(Enum):
    NETWORK = "NETWORK"
    ADJACENT_NETWORK = "ADJACENT_NETWORK"
    LOCAL = "LOCAL"
    PHYSICAL = "PHYSICAL"


class ModifiedAttackVectorType(Enum):
    NETWORK = "NETWORK"
    ADJACENT_NETWORK = "ADJACENT_NETWORK"
    LOCAL = "LOCAL"
    PHYSICAL = "PHYSICAL"
    NOT_DEFINED = "NOT_DEFINED"


class AttackComplexityType(Enum):
    HIGH = "HIGH"
    LOW = "LOW"


class ModifiedAttackComplexityType(Enum):
    HIGH = "HIGH"
    LOW = "LOW"
    NOT_DEFINED = "NOT_DEFINED"


class PrivilegesRequiredType(Enum):
    HIGH = "HIGH"
    LOW = "LOW"
    NONE = "NONE"


class ModifiedPrivilegesRequiredType(Enum):
    HIGH = "HIGH"
    LOW = "LOW"
    NONE = "NONE"
    NOT_DEFINED = "NOT_DEFINED"


class UserInteractionType(Enum):
    NONE = "NONE"
    REQUIRED = "REQUIRED"


class ModifiedUserInteractionType(Enum):
    NONE = "NONE"
    REQUIRED = "REQUIRED"
    NOT_DEFINED = "NOT_DEFINED"


class ScopeType(Enum):
    UNCHANGED = "UNCHANGED"
    CHANGED = "CHANGED"


class ModifiedScopeType(Enum):
    UNCHANGED = "UNCHANGED"
    CHANGED = "CHANGED"
    NOT_DEFINED = "NOT_DEFINED"


class CiaType(Enum):
    NONE = "NONE"
    LOW = "LOW"
    HIGH = "HIGH"


class ModifiedCiaType(Enum):
    NONE = "NONE"
    LOW = "LOW"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


class ExploitCodeMaturityType(Enum):
    UNPROVEN = "UNPROVEN"
    PROOF_OF_CONCEPT = "PROOF_OF_CONCEPT"
    FUNCTIONAL = "FUNCTIONAL"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


class RemediationLevelType(Enum):
    OFFICIAL_FIX = "OFFICIAL_FIX"
    TEMPORARY_FIX = "TEMPORARY_FIX"
    WORKAROUND = "WORKAROUND"
    UNAVAILABLE = "UNAVAILABLE"
    NOT_DEFINED = "NOT_DEFINED"


class ConfidenceType(Enum):
    UNKNOWN = "UNKNOWN"
    REASONABLE = "REASONABLE"
    CONFIRMED = "CONFIRMED"
    NOT_DEFINED = "NOT_DEFINED"


class CiaRequirementType(Enum):
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


class ScoreType(RootModel[float]):
    root: Annotated[float, Field(ge=0.0, le=10.0)]


class SeverityType(Enum):
    NONE = "NONE"
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"
    CRITICAL = "CRITICAL"


class AccessVectorType(Enum):
    NETWORK = "NETWORK"
    ADJACENT_NETWORK = "ADJACENT_NETWORK"
    LOCAL = "LOCAL"


class AccessComplexityType(Enum):
    HIGH = "HIGH"
    MEDIUM = "MEDIUM"
    LOW = "LOW"


class AuthenticationType(Enum):
    MULTIPLE = "MULTIPLE"
    SINGLE = "SINGLE"
    NONE = "NONE"


class CiaTypeModel(Enum):
    NONE = "NONE"
    PARTIAL = "PARTIAL"
    COMPLETE = "COMPLETE"


class ExploitabilityType(Enum):
    UNPROVEN = "UNPROVEN"
    PROOF_OF_CONCEPT = "PROOF_OF_CONCEPT"
    FUNCTIONAL = "FUNCTIONAL"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


class ReportConfidenceType(Enum):
    UNCONFIRMED = "UNCONFIRMED"
    UNCORROBORATED = "UNCORROBORATED"
    CONFIRMED = "CONFIRMED"
    NOT_DEFINED = "NOT_DEFINED"


class CollateralDamagePotentialType(Enum):
    NONE = "NONE"
    LOW = "LOW"
    LOW_MEDIUM = "LOW_MEDIUM"
    MEDIUM_HIGH = "MEDIUM_HIGH"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


class TargetDistributionType(Enum):
    NONE = "NONE"
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


class Type(Enum):
    finder = "finder"
    reporter = "reporter"
    analyst = "analyst"
    coordinator = "coordinator"
    remediation_developer = "remediation developer"
    remediation_reviewer = "remediation reviewer"
    remediation_verifier = "remediation verifier"
    tool = "tool"
    sponsor = "sponsor"
    other = "other"
